from redis_pydantic.types.base import RedisType, RedisSerializer


class BooleanSerializer(RedisSerializer):
    def serialize_value(self, value):
        return bool(value) if value is not None else False

    def deserialize_value(self, value):
        if isinstance(value, bool):
            return value
        elif isinstance(value, (int, float)):
            return bool(value)
        elif isinstance(value, str):
            return value.lower() in ("true", "1")
        else:
            return bool(value)


class RedisBool(int, RedisType):
    serializer = BooleanSerializer(bool, None)

    def __new__(cls, value=False, **kwargs):
        if value is None:
            value = False
        return super().__new__(cls, bool(value))

    def __init__(self, value=False, **kwargs):
        RedisType.__init__(self, **kwargs)

    async def load(self):
        redis_value = await self.client.json().get(self.redis_key, self.field_path)
        if redis_value is not None:
            return self.serializer.deserialize_value(redis_value)
        return False

    async def set(self, value: bool):
        if not isinstance(value, bool):
            raise TypeError("Value must be bool")

        serialized_value = self.serializer.serialize_value(value)
        return await self.client.json().set(
            self.redis_key, self.json_path, serialized_value
        )

    def clone(self):
        return bool(self)



from typing import TypeVar, Generic, get_args

from redis_pydantic.config import RedisFieldConfig
from redis_pydantic.types.base import GenericRedisType, RedisSerializer
from redis_pydantic.types.utils import update_keys_in_pipeline

T = TypeVar("T")


class DictSerializer(RedisSerializer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.inner_serializer = self._create_inner_serializer()

    def _create_inner_serializer(self):
        args = get_args(self.full_type)
        if len(args) >= 2:
            value_type = args[1]
            # dict[key_type, value_type] - we care about value_type
            return self.serializer_creator(value_type)
        return None

    def serialize_value(self, value):
        if self.inner_serializer:
            return {
                k: self.inner_serializer.serialize_value(v) for k, v in value.items()
            }
        return dict(value)

    def deserialize_value(self, value):
        if value is None:
            return None
        if self.inner_serializer:
            return {
                k: self.inner_serializer.deserialize_value(v) for k, v in value.items()
            }
        return dict(value)


class RedisDict(dict[str, T], GenericRedisType, Generic[T]):
    def __init__(self, value=None, *args, **kwargs):
        # Extract value if passed as keyword argument
        GenericRedisType.__init__(self, **kwargs)
        super().__init__(value, *args)

    async def load(self):
        # Get all items from Redis dict
        redis_items = await self.client.json().get(self.redis_key, self.field_path)

        if redis_items is None:
            redis_items = {}

        # Deserialize items using serializer
        deserialized_items = {}
        for key, value in redis_items.items():
            if self.serializer is not None:
                # Use serializer to deserialize the value
                deserialized_value = self.serializer.deserialize_value(value)
                deserialized_items[key] = deserialized_value
            else:
                # No type conversion
                deserialized_items[key] = value

        # Clear local dict and populate with Redis data
        super().clear()
        super().update(deserialized_items)

    async def aset_item(self, key, value):
        super().__setitem__(key, value)

        # Serialize the value for Redis storage
        serialized_value = (
            self.serializer.serialize_value(value) if self.serializer else value
        )
        return await self.client.json().set(
            self.redis_key, self.json_field_path(key), serialized_value
        )

    def __setitem__(self, key, value):
        redis_type, kwargs = self.inner_type
        field_path = f"{self.field_path}.{key}"
        new_val = self.inst_init(
            redis_type,
            value,
            self.redis_key,
            **kwargs,
            redis=self.redis,
            field_path=field_path,
            _field_config_override=RedisFieldConfig(field_path=field_path),
        )
        return super().__setitem__(key, new_val)

    async def adel_item(self, key):
        super().__delitem__(key)

        return await self.client.json().delete(
            self.redis_key, self.json_field_path(key)
        )

    def _parse_redis_json_value(self, result):
        """Parse JSON-encoded value returned from Redis Lua scripts."""
        if isinstance(result, bytes):
            result = result.decode()
        if result.startswith("[") and result.endswith("]"):
            # Remove JSON array wrapping for single values
            result = result[1:-1]
            # Handle string values that were quoted
            if result.startswith('"') and result.endswith('"'):
                result = result[1:-1]
        return result

    async def aupdate(self, **kwargs):
        self.update(**kwargs)
        redis_params = {
            self.json_field_path(key): (
                self.serializer.serialize_value(v) if self.serializer else v
            )
            for key, v in kwargs.items()
        }
        if self.pipeline:
            update_keys_in_pipeline(self.pipeline, self.redis_key, **redis_params)
            return

        async with self.redis.pipeline() as pipeline:
            update_keys_in_pipeline(pipeline, self.redis_key, **redis_params)
            await pipeline.execute()

    async def apop(self, key, default=None):
        # Redis Lua script for atomic get-and-delete operation
        pop_script = """
        local key = KEYS[1]
        local path = ARGV[1]
        local target_key = ARGV[2]
        
        -- Get the value from the JSON object
        local value = redis.call('JSON.GET', key, path .. '.' .. target_key)
        
        if value and value ~= '[]' and value ~= 'null' then
            -- Delete the key from the JSON object
            redis.call('JSON.DEL', key, path .. '.' .. target_key)
            return value
        else
            return nil
        end
        """

        # Execute the script atomically
        result = await self.client.eval(
            pop_script, 1, self.redis_key, self.json_path, key
        )

        if result is None:
            # Key doesn't exist in Redis
            if default is not None:
                return default
            else:
                raise KeyError(key)

        # Key exists in Redis, pop from local dict (it should exist there too)
        super().pop(
            key, None
        )  # Use None default to avoid KeyError if local is out of sync

        # Deserialize the value using serializer
        parsed_result = self._parse_redis_json_value(result)
        if self.serializer is not None:
            return self.serializer.deserialize_value(parsed_result)
        return parsed_result

    async def apopitem(self):
        # Redis Lua script for atomic get-arbitrary-key-and-delete operation
        popitem_script = """
        local key = KEYS[1]
        local path = ARGV[1]
        
        -- Get all the keys from the JSON object
        local keys_result = redis.call('JSON.OBJKEYS', key, path)
        
        if keys_result and type(keys_result) == 'table' and #keys_result > 0 then
            -- Get the first key from the result
            local keys = keys_result
            if type(keys[1]) == 'table' then
                keys = keys[1]  -- Sometimes Redis returns nested arrays
            end
            
            if #keys > 0 then
                local first_key = tostring(keys[1])
                -- Get the value for this key
                local value = redis.call('JSON.GET', key, path .. '.' .. first_key)
                
                if value then
                    -- Delete the key from the JSON object
                    redis.call('JSON.DEL', key, path .. '.' .. first_key)
                    return {first_key, value}
                end
            end
        end
        
        return nil
        """

        # Execute the script atomically
        result = await self.client.eval(
            popitem_script, 1, self.redis_key, self.json_path
        )

        if result is not None:
            redis_key, redis_value = result
            # Pop the same key from local dict
            super().pop(
                redis_key.decode() if isinstance(redis_key, bytes) else redis_key
            )
            parsed_value = self._parse_redis_json_value(redis_value)
            if self.serializer is not None:
                parsed_value = self.serializer.deserialize_value(parsed_value)
            return parsed_value
        else:
            # If Redis is empty but local dict has items, raise error for consistency
            raise KeyError("popitem(): dictionary is empty")

    async def aclear(self):
        # Clear local dict
        super().clear()

        # Clear Redis dict
        return await self.client.json().delete(self.redis_key, self.json_path)

    def clone(self):
        return dict.copy(self)

    @classmethod
    def find_inner_type(cls, type_):
        args = get_args(type_)
        return args[1]

    def serialize_value(self, value):
        return {k: self.serializer.serialize_value(v) for k, v in value.items()}

    def deserialize_value(self, value):
        return {k: self.serializer.deserialize_value(v) for k, v in value.items()}



from typing import TypeVar, Generic
from typing import get_args

from redis_pydantic.config import RedisFieldConfig
from redis_pydantic.types.base import GenericRedisType, RedisSerializer
from redis_pydantic.types.utils import noop

T = TypeVar("T")


class ListSerializer(RedisSerializer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.inner_serializer = self._create_inner_serializer()

    def _create_inner_serializer(self):
        args = get_args(self.full_type)
        if args:
            inner_type = args[0]
            return self.serializer_creator(inner_type)
        return None

    def serialize_value(self, value):
        if value is None:
            return None
        if self.inner_serializer:
            return [self.inner_serializer.serialize_value(v) for v in value]
        return list(value)

    def deserialize_value(self, value):
        if value is None:
            return None
        if self.inner_serializer:
            return [self.inner_serializer.deserialize_value(v) for v in value]
        return list(value)


class RedisList(list[T], GenericRedisType, Generic[T]):
    def __init__(self, value, *args, **kwargs):
        super().__init__(value)
        GenericRedisType.__init__(self, *args, **kwargs)

    async def load(self):
        # Get all items from Redis list
        redis_items = await self.client.json().get(self.redis_key, self.field_path)

        if redis_items is None:
            redis_items = []

        # Deserialize items using serializer
        deserialized_items = []
        for item in redis_items:
            if self.serializer is not None:
                deserialized_item = self.serializer.deserialize_value(item)
                deserialized_items.append(deserialized_item)
            else:
                # Fallback to decode bytes if no serializer
                decoded_item = item.decode() if isinstance(item, bytes) else item
                deserialized_items.append(decoded_item)

        # Clear local list and populate with Redis data
        super().clear()
        super().extend(deserialized_items)

    def json_field_path(self, field_name: str):
        return f"{self.json_path}[{field_name}]"

    def __setitem__(self, key, value):
        redis_type, kwargs = self.inner_type
        field_path = f"{self.field_path}[{key}]"
        new_val = self.inst_init(
            redis_type,
            value,
            self.redis_key,
            **kwargs,
            redis=self.redis,
            field_path=field_path,
            _field_config_override=RedisFieldConfig(field_path=field_path),
        )
        return super().__setitem__(key, new_val)

    async def aappend(self, __object):
        super().append(__object)

        # Serialize the object for Redis storage
        serialized_object = (
            self.serializer.serialize_value(__object) if self.serializer else __object
        )
        return await self.client.json().arrappend(
            self.redis_key, self.json_path, serialized_object
        )

    async def aextend(self, __iterable):
        items = list(__iterable)
        super().extend(items)

        # Convert iterable to list and serialize items
        if items:
            # Serialize all items for Redis storage
            serialized_items = [
                self.serializer.serialize_value(item) if self.serializer else item
                for item in items
            ]

            return await self.client.json().arrappend(
                self.redis_key,
                self.json_path,
                *serialized_items,
            )

        return await noop()

    async def apop(self, index=-1):
        super().pop(index)
        arrpop = await self.client.json().arrpop(self.redis_key, self.json_path, index)
        return (
            self.serializer.deserialize_value(arrpop[0])
            if self.serializer
            else arrpop[0]
        )

    async def ainsert(self, index, __object):
        super().insert(index, __object)

        # Serialize the object for Redis storage
        serialized_object = (
            self.serializer.serialize_value(__object) if self.serializer else __object
        )

        return await self.client.json().arrinsert(
            self.redis_key, self.json_path, index, serialized_object
        )

    async def aclear(self):
        # Clear local list
        super().clear()

        # Clear Redis list
        return await self.client.json().delete(self.redis_key, self.json_path)

    def clone(self):
        return list.copy(self)

    def serialize_value(self, value):
        return [self.serializer.serialize_value(v) for v in value]

    def deserialize_value(self, value):
        return [self.serializer.deserialize_value(v) for v in value]



import asyncio
import contextlib
import uuid
from datetime import timedelta
from typing import get_origin, ClassVar, Union, get_args, Any

from pydantic.fields import ModelPrivateAttr
from redis.asyncio import Redis


@contextlib.asynccontextmanager
async def acquire_lock(
    redis: Redis, key: str, lock_timeout: timedelta | int = 10, sleep_time: int = 0.1
):
    lock_key = f"{key}:lock"
    lock_token = str(uuid.uuid4())
    while not await redis.set(lock_key, lock_token, nx=True, ex=lock_timeout):
        await asyncio.sleep(sleep_time)
    try:
        yield
    finally:
        await redis.delete(lock_key)


def get_public_instance_annotations(cls):
    all_annotations = {}

    for parent in reversed(cls.__mro__[:-1]):
        if hasattr(parent, "__annotations__"):
            for name, annotation in parent.__annotations__.items():
                # Skip ClassVar fields
                if get_origin(annotation) is ClassVar:
                    continue

                # Skip fields with excluded types as default values
                field_value = getattr(parent, name, None)
                if isinstance(field_value, ModelPrivateAttr):
                    continue

                all_annotations[name] = annotation

    return all_annotations


def get_actual_type(annotation: Any) -> Any:
    origin = get_origin(annotation)
    if origin is Union:
        args = get_args(annotation)
        # Handle Optional[T] which is Union[T, None]
        non_none_args = [arg for arg in args if arg is not type(None)]
        if len(non_none_args) == 1:
            return non_none_args[0]
        return annotation  # Return as-is for complex Union types
    return annotation



import contextvars
from typing import Optional

from redis.asyncio.client import Redis

# Create a context variable to store the context
_context_var: contextvars.ContextVar[Optional["Redis"]] = contextvars.ContextVar(
    "redis", default=None
)



import abc
from abc import ABC
from typing import get_args, Callable

from redis.asyncio.client import Redis

from redis_pydantic.context import _context_var


class RedisSerializer(ABC):
    def __init__(self, full_type: type, serializer_creator: Callable):
        self.full_type = full_type
        self.serializer_creator = serializer_creator

    def serialize_value(self, value):
        return value

    def deserialize_value(self, value):
        return value


class RedisType(ABC):
    serializer: RedisSerializer = None

    def __init__(
        self,
        *args,
        redis_key: str = "",
        field_path: str = "",
        redis: Redis = None,
        **kwargs,
    ):
        self.redis_key = redis_key
        self.field_path = field_path
        self.redis = redis

    @property
    def pipeline(self):
        return _context_var.get()

    @property
    def client(self):
        return _context_var.get() or self.redis

    @property
    def json_path(self):
        return f"$.{self.field_path}"

    def json_field_path(self, field_name: str):
        return f"{self.json_path}.{field_name}"

    @abc.abstractmethod
    async def load(self):
        pass

    @abc.abstractmethod
    def clone(self):
        pass

    def serialize_value(self, value):
        return self.serializer.serialize_value(value)

    def deserialize_value(self, value):
        return self.serializer.deserialize_value(value)


class GenericRedisType(RedisType, ABC):
    def __init__(
        self,
        serializer_creator: Callable,
        type_creator: Callable,
        inst_init: Callable,
        full_type: type,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.serializer_creator = serializer_creator
        self.type_creator = type_creator
        self.inst_init = inst_init
        self.full_type = full_type
        inner_type = self.find_inner_type(full_type)
        self.serializer = serializer_creator(inner_type)
        self.inner_type = self.type_creator("lst", inner_type)

    @classmethod
    def find_inner_type(cls, type_):
        args = get_args(type_)
        return args[0]
